@@ -42,11 +42,21 @@
 #include "Outline.h"
 #include "UnicodeMap.h"
 #include "GlobalParams.h"
+#include "gpdf-control-private.h"
 
 BEGIN_EXTERN_C
 
 #include "gpdf-bookmarks-view.h"
 
+struct _GPdfBookmarksGenState {
+	GList *items;
+	int i;
+	GtkTreeIter *parent_iter;
+	int level; 
+	struct _GPdfBookmarksGenState *next; 
+};
+typedef struct _GPdfBookmarksGenState GPdfBookmarksGenState; 
+
 struct _GPdfBookmarksViewPrivate {
 
 	GPdfView *gpdf_view; 
@@ -80,6 +90,13 @@ struct _GPdfBookmarksViewPrivate {
 	UnicodeMap *umap;
 
 	gint current_page;
+
+	gboolean generation_terminated;
+	GPdfBookmarksGenState *generation_state, *generation_head;
+	guint idle_id; 
+	gulong idle_dcon_id;
+
+	GPdfControl *parent; 
 };
 
 static void	gpdf_bookmarks_view_class_init	      (GPdfBookmarksViewClass*); 
@@ -120,6 +137,12 @@ enum {
 	LAST_SIGNAL
 };
 
+enum {
+	PROP_0,
+	PROP_VIEW,
+	PROP_CONTROL
+};
+
 #define POPUP_MENU_EXPAND_ITEM			"Expand"
 #define POPUP_MENU_EXPAND_ITEM_TIP		"Expand all bookmarks under this one"
 #define POPUP_MENU_COLLAPSE_ITEM		"Collapse"
@@ -155,6 +178,31 @@ static guint gpdf_bookmarks_view_signals [LAST_SIGNAL];
 
 GPDF_CLASS_BOILERPLATE(GPdfBookmarksView, gpdf_bookmarks_view, GtkScrolledWindow, GTK_TYPE_SCROLLED_WINDOW); 
 
+static void
+gpdf_bookmarks_view_set_property (GObject *object, guint param_id,
+				  const GValue *value, GParamSpec *pspec)
+{
+	GPdfBookmarksView *bookmarks_view;
+
+	g_return_if_fail (GPDF_IS_BOOKMARKS_VIEW (object));
+
+	bookmarks_view = GPDF_BOOKMARKS_VIEW (object);
+
+	switch (param_id) {
+	case PROP_VIEW:
+		bookmarks_view->priv->gpdf_view = 
+		  GPDF_VIEW (g_value_get_object (value));
+		break;
+	case PROP_CONTROL:
+		bookmarks_view->priv->parent = 
+		  GPDF_CONTROL (g_value_get_object (value));
+		break;
+	default:
+		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, param_id, pspec);
+		break;
+	}
+}
+
 static void
 gpdf_bookmarks_view_class_init (GPdfBookmarksViewClass *klass)
 {
@@ -166,6 +214,25 @@ gpdf_bookmarks_view_class_init (GPdfBookmarksViewClass *klass)
 
 	gobject_class->dispose = gpdf_bookmarks_view_dispose;
 	gobject_class->finalize = gpdf_bookmarks_view_finalize;
+	gobject_class->set_property = gpdf_bookmarks_view_set_property;
+
+	g_object_class_install_property (
+		gobject_class, PROP_VIEW,
+		g_param_spec_object (
+			"view",
+			"Parent view",
+			"Parent view",
+			GPDF_TYPE_VIEW, 
+		        (GParamFlags)(G_PARAM_CONSTRUCT_ONLY | G_PARAM_WRITABLE)));
+
+	g_object_class_install_property (
+		gobject_class, PROP_CONTROL,
+		g_param_spec_object (
+			"control",
+			"Parent control",
+			"Parent control",
+			GPDF_TYPE_CONTROL, 
+		        (GParamFlags)(G_PARAM_CONSTRUCT_ONLY | G_PARAM_WRITABLE)));
 
 	gpdf_bookmarks_view_signals [BOOKMARK_SELECTED_SIGNAL] =
 	  g_signal_new (
@@ -341,6 +408,8 @@ gpdf_bookmarks_view_row_expanded_cb (GtkTreeView *treeview,
 	g_return_if_fail (GPDF_IS_NON_NULL_BOOKMARKS_VIEW (bookmarks_view));
 	priv = bookmarks_view->priv;
 
+	if (!priv->generation_terminated) return; 
+
 	/* Get bookmark page nr */
 	gtk_tree_model_get_value (GTK_TREE_MODEL(priv->model),
 				  iter,
@@ -379,6 +448,8 @@ gpdf_bookmarks_view_row_collapsed_cb (GtkTreeView *treeview,
 	g_return_if_fail (GPDF_IS_NON_NULL_BOOKMARKS_VIEW (bookmarks_view));
 	priv = bookmarks_view->priv; 
 
+	if (!priv->generation_terminated) return; 
+
 	/* Get bookmark page nr */
 	gtk_tree_model_get_value (GTK_TREE_MODEL (priv->model),
 				  iter,
@@ -486,7 +557,8 @@ gpdf_bookmarks_view_page_changed_cb (GPdfView *gpdf_view,
 	bookmarks_view->priv->current_page = page;
 
 	/* If view not yet constructed */
-	if (!bookmarks_view->priv->model) return;
+	if (!bookmarks_view->priv->model || 
+	    !bookmarks_view->priv->generation_terminated) return;
 
 	/* Else update  bookmarks tree status */
 	model = GTK_TREE_MODEL (bookmarks_view->priv->model);
@@ -496,16 +568,19 @@ gpdf_bookmarks_view_page_changed_cb (GPdfView *gpdf_view,
 }
 
 GtkWidget *
-gpdf_bookmarks_view_new (GtkWidget *gpdf_view)
+gpdf_bookmarks_view_new (GPdfControl *control, GtkWidget *gpdf_view)
 {
 	GPdfBookmarksView *bookmarks_view;
 
 	bookmarks_view =
 	  GPDF_BOOKMARKS_VIEW (g_object_new (
-		  GPDF_TYPE_BOOKMARKS_VIEW, NULL));
+		  GPDF_TYPE_BOOKMARKS_VIEW,
+		  "view", GPDF_VIEW (gpdf_view),
+		  "control", control,
+		  NULL));
 
-	bookmarks_view->priv->gpdf_view = GPDF_VIEW (gpdf_view);
 	g_object_ref (G_OBJECT (bookmarks_view->priv->gpdf_view));
+	g_object_ref (G_OBJECT (bookmarks_view->priv->parent));
 	
 	gpdf_bookmarks_view_construct (bookmarks_view);
 	
@@ -557,6 +632,10 @@ gpdf_bookmarks_view_dispose (GObject *object)
 		priv->popup_menu = NULL;
 		g_free (priv->popup_menu_uiinfo);
 	}
+	if (priv->parent) {
+		g_object_unref (priv->parent);
+		priv->parent = NULL; 
+	}
 
 	G_OBJECT_CLASS (parent_class)->dispose (object);
 }
@@ -702,7 +781,7 @@ gpdf_bookmarks_view_bookmark_select_func (GtkTreeSelection  *selection,
 			item = (OutlineItem *)g_value_peek_pointer ((const GValue*) &selection_item);
 
 			gpdf_bookmarks_view_emit_bookmark_selected (bookmarks_view,
-								    item->getAction ());
+								    item ? item->getAction () : NULL);
 		}
 		
 		ret = TRUE;
@@ -713,13 +792,10 @@ gpdf_bookmarks_view_bookmark_select_func (GtkTreeSelection  *selection,
 }
 
 static void
-gpdf_bookmarks_view_recurse_outlines (GPdfBookmarksView *bookmarks_view, 
-				      GtkTreeIter *parent_iter, 
-				      GList *items)
+gpdf_bookmarks_view_flat_recurse_outlines (GPdfBookmarksView *bookmarks_view)
 {
 	GPdfBookmarksViewPrivate *priv;
 	GdkPixbuf *bookmark_pixbuf = NULL; 
-	OutlineItem *anItem;
 	LinkAction *link_action = NULL;
 	LinkNamed *link_named = NULL;
 	LinkGoTo *link_goto;
@@ -730,119 +806,188 @@ gpdf_bookmarks_view_recurse_outlines (GPdfBookmarksView *bookmarks_view,
 	int i, page = 0;
 	gchar *pagestr = NULL;
 	gboolean is_expanded; 
+	gchar *status = NULL;
+	double ratio = 0.0; 
 	
 	g_return_if_fail (GPDF_IS_NON_NULL_BOOKMARKS_VIEW (bookmarks_view)); 
 
 	priv = bookmarks_view->priv; 
 
-	for (i = 0, anItem = (OutlineItem *)items->get(0); 
-	     i < items->getLength(); 
-	     i++, anItem = (OutlineItem *)items->get(i))
+	i = priv->generation_state->i;
+	if (i < priv->generation_state->items->getLength())
 	{
 		GtkTreeIter child_iter;
+		OutlineItem *anItem;
+		Unicode *title; 
+		
+		anItem = (OutlineItem *)priv->generation_state->items->get(i);
+		priv->generation_state->i++;
 		
-		gtk_tree_store_append (GTK_TREE_STORE (priv->model),
-				       &child_iter,
-				       parent_iter);
-
-		is_expanded =
-			(gtk_tree_view_row_expanded (
-				GTK_TREE_VIEW (priv->treeview),
-				gtk_tree_model_get_path (
-					GTK_TREE_MODEL (priv->model),
-					&child_iter)));
-
 		link_action = anItem->getAction ();
-		if (link_action == NULL) 
-		  continue;
-		switch (link_action->getKind ()) {
+		title = anItem->getTitle (); 
+		if (link_action) {
+			gtk_tree_store_append (GTK_TREE_STORE (priv->model),
+					       &child_iter,
+					       priv->generation_state->parent_iter);
+			
+			is_expanded =
+				(gtk_tree_view_row_expanded (
+					GTK_TREE_VIEW (priv->treeview),
+					gtk_tree_model_get_path (
+						GTK_TREE_MODEL (priv->model),
+						&child_iter)));
 
-		case actionGoTo:
-			link_goto = dynamic_cast <LinkGoTo *> (link_action);
-			link_dest = link_goto->getDest ();
-			named_dest = link_goto->getNamedDest (); 
-			if (link_dest != NULL) {
-				link_dest = link_dest->copy ();
-			} else if (named_dest != NULL) {
-				named_dest = named_dest->copy ();
-				link_dest = priv->pdf_doc->findDest (named_dest);
-				delete named_dest;
-			}
-			if (link_dest != NULL) {
-				if (link_dest->isPageRef ()) {
-					page_ref = link_dest->getPageRef ();
-					page = priv->pdf_doc->findPage (page_ref.num, page_ref.gen);
-				} else {
-					page = link_dest->getPageNum ();
+			switch (link_action->getKind ()) {
+
+			case actionGoTo:
+				link_goto = dynamic_cast <LinkGoTo *> (link_action);
+				link_dest = link_goto->getDest ();
+				named_dest = link_goto->getNamedDest (); 
+				if (link_dest != NULL) {
+					link_dest = link_dest->copy ();
+				} else if (named_dest != NULL) {
+					named_dest = named_dest->copy ();
+					link_dest = priv->pdf_doc->findDest (named_dest);
+					delete named_dest;
 				}
+				if (link_dest != NULL) {
+					if (link_dest->isPageRef ()) {
+						page_ref = link_dest->getPageRef ();
+						page = priv->pdf_doc->findPage (page_ref.num, page_ref.gen);
+					} else {
+						page = link_dest->getPageNum ();
+					}
+					
+					delete link_dest;
+				}
+				pagestr = g_strdup_printf ("%d", page); 
 				
-				delete link_dest;
-			}
-			pagestr = g_strdup_printf ("%d", page); 
-
-			if (page == priv->current_page) {
-				if (is_expanded)
-					bookmark_pixbuf = priv->stock_book_opened_mark_pixbuf;
-				else
-					bookmark_pixbuf = priv->stock_book_closed_mark_pixbuf;
-			}
-			else {
-				if (is_expanded)
-					bookmark_pixbuf = priv->stock_book_opened_pixbuf;
-				else
-					bookmark_pixbuf = priv->stock_book_closed_pixbuf;
-			}
-			break;
-			
-		case actionNamed:
-			dynamic_cast <LinkNamed *> (link_action) != NULL; 
-			action_name = link_named->getName ()->getCString ();
-			if (strcmp (action_name, "NextPage") == 0) {
-				page = priv->current_page +1;
-				bookmark_pixbuf = priv->stock_next_pixbuf; 
-			} else if (strcmp (action_name, "PrevPage") == 0) {
-				page = priv->current_page -1; 
-				bookmark_pixbuf = priv->stock_prev_pixbuf; 
-			} else if (strcmp (action_name, "FirstPage") == 0) {
-				page = 1; 
-				bookmark_pixbuf = priv->stock_first_pixbuf; 
-			} else if (strcmp (action_name, "LastPage") == 0) {
-				page = priv->pdf_doc->getNumPages (); 
-				bookmark_pixbuf = priv->stock_last_pixbuf; 
-			} else if (strcmp (action_name, "Close") == 0) {
-				page = -1; 
-				bookmark_pixbuf = priv->stock_close_pixbuf; 
-			} else if (strcmp (action_name, "Quit") == 0) {
-				page = -1; 
-				bookmark_pixbuf = priv->stock_quit_pixbuf; 
-			} else {
-				g_warning ("Unimplemented named link action: %s", action_name);
+				if (page == priv->current_page) {
+					if (is_expanded)
+						bookmark_pixbuf = priv->stock_book_opened_mark_pixbuf;
+					else
+						bookmark_pixbuf = priv->stock_book_closed_mark_pixbuf;
+				}
+				else {
+					if (is_expanded)
+						bookmark_pixbuf = priv->stock_book_opened_pixbuf;
+					else
+						bookmark_pixbuf = priv->stock_book_closed_pixbuf;
+				}
+				break;
+				
+			case actionNamed:
+				link_named = dynamic_cast <LinkNamed *> (link_action); 
+				action_name = link_named->getName ()->getCString ();
+				if (strcmp (action_name, "NextPage") == 0) {
+					page = priv->current_page +1;
+					bookmark_pixbuf = priv->stock_next_pixbuf; 
+				} else if (strcmp (action_name, "PrevPage") == 0) {
+					page = priv->current_page -1; 
+					bookmark_pixbuf = priv->stock_prev_pixbuf; 
+				} else if (strcmp (action_name, "FirstPage") == 0) {
+					page = 1; 
+					bookmark_pixbuf = priv->stock_first_pixbuf; 
+				} else if (strcmp (action_name, "LastPage") == 0) {
+					page = priv->pdf_doc->getNumPages (); 
+					bookmark_pixbuf = priv->stock_last_pixbuf; 
+				} else if (strcmp (action_name, "Close") == 0) {
+					page = -1; 
+					bookmark_pixbuf = priv->stock_close_pixbuf; 
+				} else if (strcmp (action_name, "Quit") == 0) {
+					page = -1; 
+					bookmark_pixbuf = priv->stock_quit_pixbuf; 
+				} else {
+					g_warning ("Unimplemented named link action: %s", action_name);
+					pagestr = g_strdup ("Unknown"); 
+				}
+				if (page == -1)
+					pagestr = g_strdup (action_name); 
+				else if (page > 0)
+					pagestr = g_strdup_printf ("%d", page);
+				break; 
+				
+			default:
 				pagestr = g_strdup ("Unknown"); 
+				g_warning ("Unkown link action type");
 			}
-			if (page == -1)
-				pagestr = g_strdup (action_name); 
-			else if (page > 0)
-				pagestr = g_strdup_printf ("%d", page);
-			break; 
-			
-		default:
-			pagestr = g_strdup ("Unknown"); 
-			g_warning ("Unkown link action type"); 
-		}
 
-		gtk_tree_store_set (GTK_TREE_STORE (priv->model), &child_iter,
-				    GPDF_BKVIEW_COLUMN1, bookmark_pixbuf, 
-				    GPDF_BKVIEW_COLUMN2, unicode_to_char (anItem, priv->umap), 
-				    GPDF_BKVIEW_COLUMN3, pagestr,
-				    GPDF_BKVIEW_COLUMN4, (gpointer)anItem,
-				    GPDF_BKVIEW_COLUMN5, page, 
-				    -1);
+			ratio = ((double)page/(double)priv->pdf_doc->getNumPages ()) /2.0; 
+			status = g_strdup_printf ("%d/%d: %2d %%",
+						  page, 
+						  priv->pdf_doc->getNumPages () *2, 
+						  (int)(ratio*100.0));
+			gpdf_control_private_set_status (priv->parent, status);
+			gpdf_control_private_set_fraction (priv->parent, ratio); 
+			g_free (status);
+			
+			gtk_tree_store_set (GTK_TREE_STORE (priv->model), &child_iter,
+					    GPDF_BKVIEW_COLUMN1, bookmark_pixbuf, 
+					    GPDF_BKVIEW_COLUMN2, unicode_to_char (anItem, priv->umap), 
+					    GPDF_BKVIEW_COLUMN3, pagestr,
+					    GPDF_BKVIEW_COLUMN4, (gpointer)anItem,
+					    GPDF_BKVIEW_COLUMN5, page, 
+					    -1);
+			g_free (pagestr);
+		}
+		/*
+		 * Title bookmark (root bookmark on page 1) have null link_action:
+		 * just goto page 1
+		 * These appears on some pdf and no bookmarks at all were displayed
+		 * before because of early recursion failure at build time in
+		 * bookmarks view.
+		 * The following else handle this case. An expecption is made in
+		 * gpdf-view to jump to page 1.
+		 */
+		else if (title) {
+			/* Ensure we are at root */
+ 			g_assert (priv->generation_state == priv->generation_head && 
+				  priv->generation_state->level == 0);
+				  
+			gtk_tree_store_append (GTK_TREE_STORE (priv->model),
+					       &child_iter,
+					       priv->generation_state->parent_iter);
+			
+			is_expanded =
+				(gtk_tree_view_row_expanded (
+					GTK_TREE_VIEW (priv->treeview),
+					gtk_tree_model_get_path (
+						GTK_TREE_MODEL (priv->model),
+						&child_iter)));
+
+			if (is_expanded)
+				bookmark_pixbuf = priv->stock_book_opened_mark_pixbuf;
+			else
+				bookmark_pixbuf = priv->stock_book_closed_mark_pixbuf;
+			
+			gtk_tree_store_set (GTK_TREE_STORE (priv->model), &child_iter,
+					    GPDF_BKVIEW_COLUMN1, bookmark_pixbuf, 
+					    GPDF_BKVIEW_COLUMN2, unicode_to_char (anItem, priv->umap), 
+					    GPDF_BKVIEW_COLUMN3, "0",
+					    GPDF_BKVIEW_COLUMN4, (gpointer)NULL,
+					    GPDF_BKVIEW_COLUMN5, 0, 
+					    -1);
+		}
+		
 		anItem->open ();
 		if (anItem->hasKids () && anItem->getKids ())
-		  gpdf_bookmarks_view_recurse_outlines (bookmarks_view, 
-							&child_iter,
-							anItem->getKids ());
-	    }
+		{
+			GPdfBookmarksGenState *next_state = priv->generation_state;
+			
+			priv->generation_state = g_new0 (GPdfBookmarksGenState, 1);
+			priv->generation_state->items = anItem->getKids ();
+			priv->generation_state->parent_iter = gtk_tree_iter_copy (&child_iter); 
+			priv->generation_state->level = next_state->level +1; 
+			priv->generation_state->next = next_state;
+		}
+	}
+	else {
+		GPdfBookmarksGenState *prev_state = priv->generation_state->next;
+
+		g_free (priv->generation_state->parent_iter); 
+		g_free (priv->generation_state);
+		priv->generation_state = prev_state; 
+	}
 }
 
 static void
@@ -858,57 +1003,69 @@ gpdf_bookmarks_view_update_bookmarks_tree (GPdfBookmarksView *bookmarks_view)
 	g_return_if_fail (GPDF_IS_NON_NULL_BOOKMARKS_VIEW (bookmarks_view));
 	
 	priv = bookmarks_view->priv;
-	items = priv->outlines->getItems(); 
+	if (priv->generation_state == NULL)
+	{
+		items = priv->outlines->getItems(); 
 	
-	if (priv->model) {
-		gtk_tree_store_clear(priv->model);
-		g_object_unref (priv->model);
-		priv->model = NULL;
-	}
+		if (priv->model) {
+			gtk_tree_store_clear(priv->model);
+			g_object_unref (priv->model);
+			priv->model = NULL;
+		}
 	
-	priv->model = gtk_tree_store_new (NUM_COLUMNS,
-					  GDK_TYPE_PIXBUF,
-					  G_TYPE_STRING,
-					  G_TYPE_STRING,
-					  G_TYPE_POINTER,
-					  G_TYPE_INT);
+		priv->model = gtk_tree_store_new (NUM_COLUMNS,
+						  GDK_TYPE_PIXBUF,
+						  G_TYPE_STRING,
+						  G_TYPE_STRING,
+						  G_TYPE_POINTER,
+						  G_TYPE_INT);
 
-	if (!priv->umap) {
-		enc = new GString("UTF-8");
-		priv->umap = globalParams->getUnicodeMap(enc);
-		priv->umap->incRefCnt (); 
-		delete enc;
+		if (!priv->umap) {
+			enc = new GString("UTF-8");
+			priv->umap = globalParams->getUnicodeMap(enc);
+			priv->umap->incRefCnt (); 
+			delete enc;
+		}
+
+		gtk_tree_view_set_model (GTK_TREE_VIEW (priv->treeview),
+					 GTK_TREE_MODEL (priv->model));
+		
+		selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (priv->treeview));
+		gtk_tree_selection_set_select_function (selection,
+							gpdf_bookmarks_view_bookmark_select_func,
+							bookmarks_view, NULL);
+		
+		gtk_tree_view_set_rules_hint (GTK_TREE_VIEW (priv->treeview), TRUE);
+		
+		renderer = gtk_cell_renderer_pixbuf_new ();
+		column = gtk_tree_view_column_new_with_attributes (NULL, renderer,
+								   "pixbuf", GPDF_BKVIEW_COLUMN1, 
+								   NULL);
+		gtk_tree_view_append_column (GTK_TREE_VIEW (priv->treeview), column);
+		
+		renderer = gtk_cell_renderer_text_new ();
+		column = gtk_tree_view_column_new_with_attributes (_("Title"), renderer,
+								   "text", GPDF_BKVIEW_COLUMN2,
+								   NULL);
+		gtk_tree_view_append_column (GTK_TREE_VIEW (priv->treeview), column);
+		
+		renderer = gtk_cell_renderer_text_new ();
+		column = gtk_tree_view_column_new_with_attributes (_("Page"), renderer,
+								   "text", GPDF_BKVIEW_COLUMN3,
+								   NULL);
+		gtk_tree_view_append_column (GTK_TREE_VIEW (priv->treeview), column);
+
+		priv->generation_state = g_new0 (GPdfBookmarksGenState, 1);
+		priv->generation_head = priv->generation_state; 
+		priv->generation_state->items = items;
+		priv->generation_state->level = 0; 
 	}
 
-	gpdf_bookmarks_view_recurse_outlines (bookmarks_view, NULL, items); 
-
-	gtk_tree_view_set_model (GTK_TREE_VIEW (priv->treeview),
-				 GTK_TREE_MODEL (priv->model));
-	
-	selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (priv->treeview));
-	gtk_tree_selection_set_select_function (selection,
-			gpdf_bookmarks_view_bookmark_select_func,
-			bookmarks_view, NULL);
-
-	gtk_tree_view_set_rules_hint (GTK_TREE_VIEW (priv->treeview), TRUE);
-
-	renderer = gtk_cell_renderer_pixbuf_new ();
-	column = gtk_tree_view_column_new_with_attributes (NULL, renderer,
-							   "pixbuf", GPDF_BKVIEW_COLUMN1, 
-							   NULL);
-	gtk_tree_view_append_column (GTK_TREE_VIEW (priv->treeview), column);
-
-	renderer = gtk_cell_renderer_text_new ();
-	column = gtk_tree_view_column_new_with_attributes (_("Title"), renderer,
-							   "text", GPDF_BKVIEW_COLUMN2,
-							   NULL);
-	gtk_tree_view_append_column (GTK_TREE_VIEW (priv->treeview), column);
-	
-	renderer = gtk_cell_renderer_text_new ();
-	column = gtk_tree_view_column_new_with_attributes (_("Page"), renderer,
-							   "text", GPDF_BKVIEW_COLUMN3,
-							   NULL);
-	gtk_tree_view_append_column (GTK_TREE_VIEW (priv->treeview), column);
+	if (priv->generation_state == priv->generation_head && \
+	    priv->generation_head->i == priv->generation_head->items->getLength())
+		priv->generation_terminated = TRUE;
+	else
+		gpdf_bookmarks_view_flat_recurse_outlines (bookmarks_view); 
 }
 
 static void
@@ -918,6 +1075,8 @@ gpdf_bookmarks_view_selection_changed_cb (GtkTreeSelection *selection, gpointer
 
 	g_return_if_fail (GPDF_IS_NON_NULL_BOOKMARKS_VIEW (bookmarks_view));
 	
+	if (!bookmarks_view->priv->generation_terminated) return;
+		
 	gpdf_bookmarks_view_update_popup_actions (bookmarks_view);
 }
 
@@ -959,21 +1118,55 @@ gpdf_bookmarks_view_populate_idle (GPdfBookmarksView *bookmarks_view)
 	if (gpdf_bookmarks_view_have_outline_items (bookmarks_view)) {
 		
 		gpdf_bookmarks_view_update_bookmarks_tree (bookmarks_view);
-		
-		/* Add listener for popup actions sensitivity */
-		selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (priv->treeview)); 
-		priv->selection_id =
-		  g_signal_connect (G_OBJECT (selection), "changed",
-				    G_CALLBACK (gpdf_bookmarks_view_selection_changed_cb), 
-				    bookmarks_view);
-		
-		/* If root is unique, let's expand it */
-		gpdf_bookmarks_view_expand_unique_root (bookmarks_view);
+
+		if (priv->generation_terminated)
+		{
+			/* Add listener for popup actions sensitivity */
+			selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (priv->treeview)); 
+			priv->selection_id =
+				g_signal_connect (G_OBJECT (selection), "changed",
+						  G_CALLBACK (gpdf_bookmarks_view_selection_changed_cb), 
+						  bookmarks_view);
+			
+			/* If root is unique, let's expand it */
+			gpdf_bookmarks_view_expand_unique_root (bookmarks_view);
+		}
+	}
+	else {
+		priv->generation_terminated = TRUE; 
 	}
 
-	gpdf_bookmarks_view_emit_ready (bookmarks_view);
+	if (priv->generation_terminated) {
+		GdkDisplay *display;
+		GdkWindow *parent_window;
 
-	return FALSE;
+		display = gtk_widget_get_display (priv->treeview);
+		parent_window = gtk_widget_get_parent_window (priv->treeview);
+		if (GDK_IS_WINDOW (parent_window))
+			gdk_window_set_cursor (parent_window, NULL); 
+		gdk_flush();
+		g_signal_handler_disconnect (G_OBJECT (priv->parent), priv->idle_dcon_id); 
+		gpdf_control_private_clear_stack (priv->parent);
+		gpdf_control_private_set_fraction (priv->parent, 0.0); 
+		gpdf_bookmarks_view_emit_ready (bookmarks_view);
+	}
+	else {
+		GdkDisplay *display;
+		GdkCursor *cursor;
+		GdkWindow *parent_window;
+		
+		/* Set watch cursor while view not ready */
+		display = gtk_widget_get_display (priv->treeview);
+		cursor = gdk_cursor_new_for_display (display, GDK_WATCH);
+		parent_window = gtk_widget_get_parent_window (priv->treeview);
+		if (GDK_IS_WINDOW (parent_window))
+			gdk_window_set_cursor (parent_window, cursor); 
+		gdk_cursor_unref (cursor);
+		gdk_flush();
+
+	}
+
+	return (!priv->generation_terminated);
 }
 
 void
@@ -987,13 +1180,17 @@ gpdf_bookmarks_view_set_pdf_doc (GPdfBookmarksView *bookmarks_view,
 	priv = bookmarks_view->priv;
 	
 	if (pdf_doc != priv->pdf_doc)
-	{
+	{		
 		priv->pdf_doc = pdf_doc;
 		priv->outlines = pdf_doc->getOutline ();
 		/* Current page set through 'page_changed' signal */
 		
-		g_idle_add ((GSourceFunc)gpdf_bookmarks_view_populate_idle,
-			    bookmarks_view);
+		priv->idle_id = g_idle_add ((GSourceFunc)gpdf_bookmarks_view_populate_idle,
+					    bookmarks_view);
+
+		g_signal_connect_swapped (G_OBJECT (priv->parent), "disconnected", 
+					  G_CALLBACK (g_source_remove),
+					  (gpointer)priv->idle_id);
 	}
 }
 
