@@ -17,6 +17,7 @@
 #include <time.h>
 #include <unistd.h>
 #include "gpdf-control.h"
+#include "gpdf-control-private.h"
 #include "gpdf-stock-icons.h"
 #include "gpdf-view.h"
 #include "gpdf-bookmarks-view.h"
@@ -422,17 +423,31 @@ gpdf_control_private_get_annots_view (GPdfControl * control)
 }
 #endif
 
+void
+destroy_widget_unref_control (GtkWidget *widget, gint arg1, BonoboControl *control)
+{
+	g_return_if_fail (GTK_IS_WIDGET (widget));
+	g_return_if_fail (BONOBO_IS_CONTROL (control));
+
+	gtk_widget_destroy (widget);
+	bonobo_object_unref (BONOBO_OBJECT (control));
+}
+
 /* Create a HIG compliant dialog, display it and
    wait until response */
 void
 gpdf_control_private_error_dialog (GPdfControl * control,
 				   const gchar *header_text,
 				   const gchar *body_text,
-				   gboolean modal)
+				   gboolean modal,
+				   gboolean ref_parent)
 {
 	GtkWidget *dlg, *but;
 
 	g_return_if_fail (GPDF_IS_NON_NULL_CONTROL (control));
+	if (ref_parent)
+		bonobo_object_ref (BONOBO_OBJECT (control));
+
 	dlg = gpdf_hig_dialog_new (GTK_STOCK_DIALOG_ERROR,
 				   header_text,
 				   body_text,
@@ -446,10 +461,14 @@ gpdf_control_private_error_dialog (GPdfControl * control,
 	if (modal) {
 		gtk_dialog_run (GTK_DIALOG (dlg));
 		gtk_widget_destroy (dlg);
+		if (ref_parent)
+			bonobo_object_unref (BONOBO_OBJECT (control));
 	} else {
 		g_signal_connect (G_OBJECT (dlg), "response",
-				  G_CALLBACK (gtk_widget_destroy),
-				  NULL);
+				  (ref_parent
+				   ? G_CALLBACK (destroy_widget_unref_control)
+				   : G_CALLBACK (gtk_widget_destroy)),
+				  control);
 		gtk_widget_show (dlg);
 	}		
 }
@@ -535,6 +554,7 @@ gpdf_control_private_display_help (BonoboControl *control,
 		gpdf_control_private_error_dialog (gpdf_control,
 						   _("Documentation Error"),
 						   msg,
+						   FALSE,
 						   FALSE);
 	}
 }
@@ -873,6 +893,7 @@ gpdf_control_no_postscript_printer_alert (GPdfControl *control,
 		control,
 		_("Printing is not supported on this printer."),
 		secondary,
+		FALSE,
 		FALSE);
 
 	g_free (secondary);
@@ -1010,6 +1031,7 @@ static gchar *
 gpdf_control_get_filename (BonoboControl *control)
 {
 	GtkWidget *filechooser;
+	GtkFileFilter *pdf_filter, *all_filter;
 	gchar *pathname = NULL; 
 
 	filechooser = gtk_file_chooser_dialog_new (_("Save a Copy"),
@@ -1019,6 +1041,15 @@ gpdf_control_get_filename (BonoboControl *control)
 						   GTK_STOCK_SAVE, GTK_RESPONSE_OK,
 						   NULL);
 	gtk_window_set_modal (GTK_WINDOW (filechooser), TRUE);
+	pdf_filter = gtk_file_filter_new ();
+	gtk_file_filter_set_name (pdf_filter, _("PDF Documents"));
+	gtk_file_filter_add_mime_type (pdf_filter, "application/pdf");
+	gtk_file_chooser_add_filter (GTK_FILE_CHOOSER (filechooser), pdf_filter);
+	all_filter = gtk_file_filter_new ();
+	gtk_file_filter_set_name (all_filter, _("All Files"));
+	gtk_file_filter_add_pattern (all_filter, "*");
+	gtk_file_chooser_add_filter (GTK_FILE_CHOOSER (filechooser), all_filter);
+	gtk_file_chooser_set_filter (GTK_FILE_CHOOSER (filechooser), pdf_filter);
 	
 	gtk_dialog_set_default_response (GTK_DIALOG (filechooser), GTK_RESPONSE_OK);
 
@@ -1695,7 +1726,7 @@ gpdf_control_get_bookmarks_view (GPdfControl *gpdf_control)
 	if (!priv->gpdf_bookmarks_view)
 	{
 	    priv->gpdf_bookmarks_view =
-	      gpdf_bookmarks_view_new (priv->gpdf_view);
+	      gpdf_bookmarks_view_new (gpdf_control, priv->gpdf_view);
 	    g_object_ref (G_OBJECT (priv->gpdf_bookmarks_view)); 
 	    gtk_widget_show (priv->gpdf_bookmarks_view);
 	    
@@ -1916,7 +1947,7 @@ persist_file_loading_failed_cb (GPdfPersistFile *persist_file,
 		_("Loading of %s failed."),
 		gpdf_persist_file_get_current_uri (persist_file));
 
-	gpdf_control_private_error_dialog (control, primary, message, FALSE);
+	gpdf_control_private_error_dialog (control, primary, message, FALSE, TRUE);
 
 	g_free (primary);
 
@@ -2294,6 +2325,12 @@ gpdf_control_setup_view_widgets (GPdfControl *control)
 	control->priv->gpdf_view = gpdf_view;
 	gtk_container_add (GTK_CONTAINER (sw), gpdf_view);
 
+	/* Workaround for 132489 */
+	gtk_scrolled_window_get_hadjustment (GTK_SCROLLED_WINDOW (sw))
+		->step_increment = 10;
+	gtk_scrolled_window_get_vadjustment (GTK_SCROLLED_WINDOW (sw))
+		->step_increment = 10;
+	
 	g_signal_connect (G_OBJECT (gpdf_view), "page_changed",
 			  G_CALLBACK (gc_page_changed_enable_page_buttons),
 			  control);
