@@ -61,7 +61,7 @@
 #define GPDF_ADD_STOCK_ICON(id, file, def_id)				       \
 {				  					       \
 	GdkPixbuf *pixbuf;						       \
-	GtkIconSet *icon_set;						       \
+	GtkIconSet *icon_set = NULL;					       \
         pixbuf = gdk_pixbuf_new_from_file (GNOMEICONDIR "/gpdf/" file, NULL);  \
         if (pixbuf) {							       \
         	icon_set = gtk_icon_set_new_from_pixbuf (pixbuf);	       \
