@@ -47,6 +47,7 @@
 #include "GfxState.h"
 #include "Thumb.h"
 #include "gpdf-view.h"
+#include "gpdf-control-private.h"
 
 BEGIN_EXTERN_C
 
@@ -92,7 +93,10 @@ struct _GPdfThumbnailsViewPrivate {
 	GtkWidget *popup_menu;
 	
 	gint current_page;
-	double zoom; 
+	double zoom;
+	gint current_gen_thumb;
+	guint idle_id; 
+	gulong idle_dcon_id; 
 
 	gint clicked_x, clicked_y;
 
@@ -135,6 +139,12 @@ enum {
 	LAST_SIGNAL
 };
 
+enum {
+	PROP_0,
+	PROP_VIEW,
+	PROP_CONTROL
+};
+
 #define POPUP_MENU_GENERATE_ITEM		"Generate Thumbnail"
 #define POPUP_MENU_GENERATE_ITEM_TIP		"Generate Thumbnail image for current page"
 #define POPUP_MENU_GENERATE_ALL_ITEM		"Generate All Thumbnails"
@@ -160,6 +170,31 @@ static guint gpdf_thumbnails_view_signals [LAST_SIGNAL];
 
 GPDF_CLASS_BOILERPLATE(GPdfThumbnailsView, gpdf_thumbnails_view, GtkScrolledWindow, GTK_TYPE_SCROLLED_WINDOW); 
 
+static void
+gpdf_thumbnails_view_set_property (GObject *object, guint param_id,
+				  const GValue *value, GParamSpec *pspec)
+{
+	GPdfThumbnailsView *thumbnails_view;
+
+	g_return_if_fail (GPDF_IS_THUMBNAILS_VIEW (object));
+
+	thumbnails_view = GPDF_THUMBNAILS_VIEW (object);
+
+	switch (param_id) {
+	case PROP_VIEW:
+		thumbnails_view->priv->gpdf_view = 
+		  GPDF_VIEW (g_value_get_object (value));
+		break;
+	case PROP_CONTROL:
+		thumbnails_view->priv->parent = 
+		  GPDF_CONTROL (g_value_get_object (value));
+		break;
+	default:
+		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, param_id, pspec);
+		break;
+	}
+}
+
 static void
 gpdf_thumbnails_view_class_init (GPdfThumbnailsViewClass *klass)
 {
@@ -171,6 +206,25 @@ gpdf_thumbnails_view_class_init (GPdfThumbnailsViewClass *klass)
 
 	gobject_class->dispose = gpdf_thumbnails_view_dispose;
 	gobject_class->finalize = gpdf_thumbnails_view_finalize;
+	gobject_class->set_property = gpdf_thumbnails_view_set_property;
+
+	g_object_class_install_property (
+		gobject_class, PROP_VIEW,
+		g_param_spec_object (
+			"view",
+			"Parent view",
+			"Parent view",
+			GPDF_TYPE_VIEW, 
+		        (GParamFlags)(G_PARAM_CONSTRUCT_ONLY | G_PARAM_WRITABLE)));
+
+	g_object_class_install_property (
+		gobject_class, PROP_CONTROL,
+		g_param_spec_object (
+			"control",
+			"Parent control",
+			"Parent control",
+			GPDF_TYPE_CONTROL, 
+		        (GParamFlags)(G_PARAM_CONSTRUCT_ONLY | G_PARAM_WRITABLE)));
 
 	gpdf_thumbnails_view_signals [THUMBNAIL_SELECTED_SIGNAL] =
 	  g_signal_new (
@@ -272,6 +326,12 @@ gpdf_thumbnails_view_page_box_coord(GPdfThumbnailsView *thumbnails_view,
 	real_width = gdk_pixbuf_get_width (priv->thumbs[page -1].pixbuf); 
 	real_height = gdk_pixbuf_get_height (priv->thumbs[page -1].pixbuf); 
 
+	/* Update page size if guess was wrong */
+	if (priv->max_page_width < real_width)
+		priv->max_page_width = real_width; 
+	if (priv->max_page_height < real_height)
+		priv->max_page_height = real_height; 
+
 	if (x) {
 		*x = GPDF_THUMBNAILS_MARGIN_LEFT +
 			((page -1) % priv->page_per_row) *
@@ -280,7 +340,6 @@ gpdf_thumbnails_view_page_box_coord(GPdfThumbnailsView *thumbnails_view,
 			 GPDF_THUMBNAILS_INNER_MARGIN_LEFT +
 			 GPDF_THUMBNAILS_SHADOW_WIDTH);
 
-		g_assert (priv->max_page_width >= real_width); 
 		*x += (priv->max_page_width -real_width)/2;
 	}
 
@@ -345,7 +404,7 @@ gpdf_thumbnails_view_set_scroll_region(GPdfThumbnailsView *thumbnails_view)
 
 	priv = thumbnails_view->priv;
 
-	for (page = 1; page <= priv->num_pages; page++) {
+	for (page = 1; page <= priv->current_gen_thumb; page++) {
 		gpdf_thumbnails_view_page_box_coord (thumbnails_view,
 						     page,
 						     &x, &y,
@@ -568,11 +627,13 @@ gpdf_thumbnails_view_new (GPdfControl *parent, GtkWidget *gpdf_view)
 
 	thumbnails_view =
 	  GPDF_THUMBNAILS_VIEW (g_object_new (
-		  GPDF_TYPE_THUMBNAILS_VIEW, NULL));
+		  GPDF_TYPE_THUMBNAILS_VIEW, 
+		  "view", GPDF_VIEW (gpdf_view),
+		  "control", GPDF_CONTROL (parent), 
+		  NULL));
 
-	thumbnails_view->priv->parent = parent;
-	thumbnails_view->priv->gpdf_view = GPDF_VIEW (gpdf_view);
 	g_object_ref (G_OBJECT (thumbnails_view->priv->gpdf_view)); 
+	g_object_ref (G_OBJECT (thumbnails_view->priv->parent)); 
 	gpdf_thumbnails_view_construct (thumbnails_view); 
 	
 	g_signal_connect (G_OBJECT (gpdf_view), "page_changed",
@@ -613,6 +674,10 @@ gpdf_thumbnails_view_dispose (GObject *object)
 		priv->popup_menu = NULL;
 		g_free (priv->popup_menu_uiinfo);
 	}
+	if (priv->parent) {
+		g_object_unref (priv->parent);
+		priv->parent = NULL; 
+	}
 
 	G_OBJECT_CLASS (parent_class)->dispose (object);
 }
@@ -978,6 +1043,7 @@ gpdf_thumbnails_view_create_thumbnail (GPdfThumbnailsView *thumbnails_view,
 	GPdfThumbnailsViewPrivate *priv; 
 	GnomeCanvasItem *group;
 	GnomeCanvasItem *item;
+	double dx, dy, dw, dh;
 	int width, height; 
 	double outerwidth, outerheight;
 	double shadow_x, shadow_y;
@@ -989,17 +1055,28 @@ gpdf_thumbnails_view_create_thumbnail (GPdfThumbnailsView *thumbnails_view,
 	priv = thumbnails_view->priv;
 
 	thumbnail->pixbuf = pixbuf;
-	
-	group = gnome_canvas_item_new(priv->root,
-				      gnome_canvas_group_get_type(),
-				      "x", (double)0,
-				      "y", (double)0, 
-				      NULL);
-	thumbnail->group = group;
 
 	width = gdk_pixbuf_get_width (pixbuf); 
 	height = gdk_pixbuf_get_height (pixbuf); 
 
+	/*
+	 * Just a guess for now, in order to have
+	 * an approximate layout while creating
+	 * thumbs
+	 */
+	dw = (double)width;
+	dh = (double)height;
+	gpdf_thumbnails_view_page_box_coord(thumbnails_view,
+					    page, 
+					    &dx, &dy,
+					    &dw, &dh);
+	group = gnome_canvas_item_new(priv->root,
+				      gnome_canvas_group_get_type(),
+				      "x", dx,
+				      "y", dy, 
+				      NULL);
+	thumbnail->group = group;
+
 	outerwidth = width +
 	  	GPDF_THUMBNAILS_INNER_MARGIN_LEFT +
 		GPDF_THUMBNAILS_INNER_MARGIN_RIGHT +
@@ -1170,9 +1247,9 @@ gpdf_thumbnails_view_render_thumbnail_image (GPdfThumbnailsView *thumbnails_view
 	pixbuf = gdk_pixbuf_new_from_data(data,
 					  GDK_COLORSPACE_RGB,
 					  FALSE,
-					  bitsPerComponent,
+					  8,
 					  width, height,
-					  width*3,
+					  width * 3,
 					  NULL, NULL);
 
 	/* Create the thumbnail itself */
@@ -1190,106 +1267,136 @@ static gboolean
 gpdf_thumbnails_view_populate_idle (GPdfThumbnailsView *thumbnails_view)
 {
 	GPdfThumbnailsViewPrivate *priv;
-	guint page;
 	Page *the_page;
 	Object the_thumb;
 	Thumb *thumb;
-	XRef *xref;
+	gboolean have_ethumbs = FALSE;
+	gchar *status = NULL;
+	double ratio = 0.0; 
 	
 	g_return_val_if_fail (GPDF_IS_NON_NULL_THUMBNAILS_VIEW (thumbnails_view),
 			      TRUE);
 	
 	priv = thumbnails_view->priv;
+	ratio = 0.5 + ((double)priv->current_gen_thumb/(double)priv->num_pages) /2.0; 
+	status = g_strdup_printf ("%d/%d: %3d %%",
+				  priv->num_pages + priv->current_gen_thumb,
+				  priv->num_pages *2, 
+				  (int)(ratio*100.0));
+	gpdf_control_private_set_status (priv->parent, status);
+	gpdf_control_private_set_fraction (priv->parent, ratio); 
+	g_free (status); 
 
-	/* Get number of pages */
-	priv->num_pages = priv->pdf_doc->getNumPages ();
-
-	/* Allocate thumbs table */
-	priv->thumbs =(GPdfThumbnail *)g_new0 (GPdfThumbnail, priv->num_pages);
-
-	/* Init page max sizes */
-	priv->max_page_width = 0;
-	priv->max_page_height = 0; 	
-
-	/* Create/Read thumbnails */
-	for (page = 1; page <= priv->num_pages; page++) {
-
-		/* Get the PDF objects for fetching thumb */
-		the_page = priv->pdf_doc->getCatalog ()->getPage (page);
-		xref = priv->pdf_doc->getXRef ();
-		the_page->getThumb(&the_thumb);
-
+	/* Get the PDF objects for fetching thumb */
+	the_page = priv->pdf_doc->getCatalog ()->getPage (++(priv->current_gen_thumb));
+	the_page->getThumb(&the_thumb);
+	
+	/*
+	 * If no thumb object available in PDF, render each page in
+	 * thumbnail mode, else fetch thumbnail object and render image
+	 * data.
+	 */
+	if (!(the_thumb.isNull () || the_thumb.isNone())) {
+		/* Build the thumbnail object */
+		thumb = new Thumb(priv->pdf_doc->getXRef (),
+				  &the_thumb);
+		
+		have_ethumbs = thumb->ok();
+	}
+	
+	if (have_ethumbs) {
+		/* Keep maximum width/height to compute layout later */
+		priv->max_page_width = MAX ((guint)thumb->getWidth(), priv->max_page_width);
+		priv->max_page_height = MAX ((guint)thumb->getHeight(), priv->max_page_height);
+		
+		gpdf_thumbnails_view_render_thumbnail_image (thumbnails_view,
+							     priv->current_gen_thumb, 
+							     thumb->getWidth(),
+							     thumb->getHeight(),
+							     thumb->getColorSpace(),
+							     thumb->getBitsPerComponent(),
+							     thumb);
+	}
+	else {
+		GtkWidget *item;
+		guint width, height;
+		
 		/*
-		 * If no thumb object available in PDF, render each page in
-		 * thumbnail mode, else fetch thumbnail object and render image
-		 * data.
+		 * FIXME: Could perhaps render only displayed thumbs. This process may
+		 * be time consuming.
 		 */
-		if (the_thumb.isNull () || the_thumb.isNone()) {
-			GtkWidget *item;
-			guint width, height;
-			
-			/*
-			 * FIXME: Could perhaps render only displayed thumbs. This process may
-			 * be time consuming.
-			 */
-			gpdf_thumbnails_view_create_thumbnail_image (thumbnails_view, the_page, page,
-								     (gint*)&width, (gint*)&height);
-
-			/* Keep maximum width/height to compute layout later */
-			priv->max_page_width = MAX (width, priv->max_page_width);
-			priv->max_page_height = MAX (height, priv->max_page_height);
-
+		gpdf_thumbnails_view_create_thumbnail_image (thumbnails_view, the_page,
+							     priv->current_gen_thumb, 
+							     (gint*)&width, (gint*)&height);
+		
+		/* Keep maximum width/height to compute layout later */
+		priv->max_page_width = MAX (width, priv->max_page_width);
+		priv->max_page_height = MAX (height, priv->max_page_height);
+		
 #if CAN_GENERATE_THUMBNAILS
-			item = priv->popup_menu_uiinfo[POPUP_MENU_GENERATE_INDEX].widget;
-			if (GTK_IS_WIDGET (item))
-			  gtk_widget_set_sensitive (item, TRUE); 
-			item = priv->popup_menu_uiinfo[POPUP_MENU_GENERATE_ALL_INDEX].widget;
-			if (GTK_IS_WIDGET (item))
-			  gtk_widget_set_sensitive (item, TRUE); 
+		item = priv->popup_menu_uiinfo[POPUP_MENU_GENERATE_INDEX].widget;
+		if (GTK_IS_WIDGET (item))
+		  gtk_widget_set_sensitive (item, TRUE); 
+		item = priv->popup_menu_uiinfo[POPUP_MENU_GENERATE_ALL_INDEX].widget;
+		if (GTK_IS_WIDGET (item))
+		  gtk_widget_set_sensitive (item, TRUE); 
 #endif
-		}
-		else {
-			/* Build the thumbnail object */
-			thumb = new Thumb(priv->pdf_doc->getXRef (),
-					  &the_thumb);
-
-			/* Keep maximum width/height to compute layout later */
-			priv->max_page_width = MAX ((guint)thumb->getWidth(), priv->max_page_width);
-			priv->max_page_height = MAX ((guint)thumb->getHeight(), priv->max_page_height);
-			
-			gpdf_thumbnails_view_render_thumbnail_image (thumbnails_view, page,
-								     thumb->getWidth(), thumb->getHeight(),
-								     thumb->getColorSpace(),
-								     thumb->getBitsPerComponent(),
-								     thumb);
-		}
 	}
 
-	/* Now width and height are known, set scroll region and relayout */
+	/* Width and height are best known, set scroll region and relayout */
 	gpdf_thumbnails_view_set_scroll_region(thumbnails_view);
-	
-	/* All thumbs are setup, let listen page canvas adjustments to update page outline */
-	g_signal_connect (GTK_LAYOUT (priv->gpdf_view)->hadjustment, "value_changed",
-			  G_CALLBACK (gpdf_thumbnails_view_page_layout_adjustment_changed),
-			  thumbnails_view); 
-	g_signal_connect (GTK_LAYOUT (priv->gpdf_view)->hadjustment, "changed",
-			  G_CALLBACK (gpdf_thumbnails_view_page_layout_adjustment_changed),
-			  thumbnails_view); 
-	g_signal_connect (GTK_LAYOUT (priv->gpdf_view)->vadjustment, "value_changed",
-			  G_CALLBACK (gpdf_thumbnails_view_page_layout_adjustment_changed),
-			  thumbnails_view); 
-	g_signal_connect (GTK_LAYOUT (priv->gpdf_view)->vadjustment, "changed",
-			  G_CALLBACK (gpdf_thumbnails_view_page_layout_adjustment_changed),
-			  thumbnails_view); 
-	
-	/* Select first page */
-	gpdf_thumbnails_view_set_selected_page (thumbnails_view, 1);
+		
+	if (priv->current_gen_thumb == priv->num_pages)
+	{
+		GdkDisplay *display;
+		GdkWindow *parent_window;
+		
+		/* All thumbs are setup, let listen page canvas adjustments to update page outline */
+		g_signal_connect (GTK_LAYOUT (priv->gpdf_view)->hadjustment, "value_changed",
+				  G_CALLBACK (gpdf_thumbnails_view_page_layout_adjustment_changed),
+				  thumbnails_view); 
+		g_signal_connect (GTK_LAYOUT (priv->gpdf_view)->hadjustment, "changed",
+				  G_CALLBACK (gpdf_thumbnails_view_page_layout_adjustment_changed),
+				  thumbnails_view); 
+		g_signal_connect (GTK_LAYOUT (priv->gpdf_view)->vadjustment, "value_changed",
+				  G_CALLBACK (gpdf_thumbnails_view_page_layout_adjustment_changed),
+				  thumbnails_view); 
+		g_signal_connect (GTK_LAYOUT (priv->gpdf_view)->vadjustment, "changed",
+				  G_CALLBACK (gpdf_thumbnails_view_page_layout_adjustment_changed),
+				  thumbnails_view); 
+		
+		/* Select first page */
+		gpdf_thumbnails_view_set_selected_page (thumbnails_view, 1);
+		
+		priv->thumbs_initialized = TRUE;
+		
+		display = gtk_widget_get_display (priv->canvas_widget);
+		parent_window = gtk_widget_get_parent_window (priv->canvas_widget);
+		if (GDK_IS_WINDOW (parent_window))
+			gdk_window_set_cursor (parent_window, NULL); 
+		gdk_flush();
 
-	priv->thumbs_initialized = TRUE;
+		gpdf_control_private_clear_stack (priv->parent);
+		gpdf_control_private_set_fraction (priv->parent, 0.0); 
+		g_signal_handler_disconnect (G_OBJECT (priv->parent), priv->idle_dcon_id); 
+		gpdf_thumbnails_view_emit_ready (thumbnails_view);
+	}
+	else {
+		GdkDisplay *display;
+		GdkCursor *cursor;
+		GdkWindow *parent_window;
+		
+		/* Set watch cursor while view not ready */
+		display = gtk_widget_get_display (priv->canvas_widget);
+		cursor = gdk_cursor_new_for_display (display, GDK_WATCH);
+		parent_window = gtk_widget_get_parent_window (priv->canvas_widget);
+		if (GDK_IS_WINDOW (parent_window))
+			gdk_window_set_cursor (parent_window, cursor); 
+		gdk_cursor_unref (cursor);
+		gdk_flush();
+	}
 
-	gpdf_thumbnails_view_emit_ready (thumbnails_view);
-
-	return FALSE; 
+	return (priv->current_gen_thumb != priv->num_pages); 
 }
 
 void
@@ -1306,12 +1413,36 @@ gpdf_thumbnails_view_set_pdf_doc (GPdfThumbnailsView *thumbnails_view,
     {
 	    priv->pdf_doc = pdf_doc;
 
+	    /* Get number of pages */
+	    priv->num_pages = priv->pdf_doc->getNumPages ();
+
+	    /* Allocate thumbs table */
+	    priv->thumbs =(GPdfThumbnail *)g_new0 (GPdfThumbnail, priv->num_pages);
+
+	    /*
+	     * Init page max sizes: We try a guess in order to have thumbs created
+	     * in an approximate place. When READY signal will be emitted the whole
+	     * layout will be recomputed with exact values.
+	     */
+	    priv->max_page_width = 45;
+	    priv->max_page_height = 90;
+
+	    /* Init index of last generated thumb */
+	    priv->current_gen_thumb = 0;
+
 	    /*
 	     * As thumb computation may be a long process
 	     * we make it as backgroud task.
 	     */
-	    g_idle_add ((GSourceFunc)gpdf_thumbnails_view_populate_idle,
-			thumbnails_view);
+	    priv->idle_id = g_idle_add_full (G_PRIORITY_DEFAULT_IDLE +1,
+					     (GSourceFunc)gpdf_thumbnails_view_populate_idle,
+					     thumbnails_view, 
+					     NULL);
+
+	    priv->idle_dcon_id =
+		    g_signal_connect_swapped (G_OBJECT (priv->parent), "disconnected",
+					      G_CALLBACK (g_source_remove),
+					      (gpointer)priv->idle_id);
     }
 }
 
