@@ -327,7 +327,7 @@ gpdf_view_link_action_named (GPdfView *gpdf_view, LinkNamed *link_named)
 static void
 gpdf_view_link_action_uri (GPdfView *gpdf_view, LinkURI *link_uri)
 {
-	GString *uri; 
+	gchar *uri; 
 	GnomeVFSMimeAction *action; 
 	GnomeVFSMimeApplication *app;
 	GnomeVFSURI *vfsuri; 
@@ -339,19 +339,15 @@ gpdf_view_link_action_uri (GPdfView *gpdf_view, LinkURI *link_uri)
 	g_return_if_fail (GPDF_IS_VIEW_WITH_DOC (gpdf_view));
 	g_return_if_fail (dynamic_cast <LinkURI *> (link_uri) != NULL);
 
-	uri = link_uri->getURI (); 
-	g_message ("gpdf_view_link_action_uri: uri = '%s'", (gchar *)uri->getCString ()); 
-	uris = g_list_append (uris, (gchar *)uri->getCString ());
+	uri = g_strdup (link_uri->getURI ()->getCString ()); 
+	uris = g_list_append (uris, uri);
 
 	vfsuri = gnome_vfs_uri_new ((const gchar *)uris->data); 
 	mime = gnome_vfs_get_mime_type_common (vfsuri);
+	gnome_vfs_uri_unref (vfsuri); 
 
-	g_message ("gpdf_view_link_action_uri: mime = '%s'", mime); 
-
-	if (!mime) {
-		mime = "text/html";
-		g_message ("gpdf_view_link_action_uri: using default mime = '%s'", mime); 
-	}
+	if (!mime)
+	  mime = "text/html";
 
 	action = gnome_vfs_mime_get_default_action (mime); 
 	if (gnome_vfs_mime_action_launch (action, uris) != GNOME_VFS_OK)
@@ -375,7 +371,8 @@ gpdf_view_link_action_uri (GPdfView *gpdf_view, LinkURI *link_uri)
 					gpdf_control_private_error_dialog (gpdf_view->priv->parent, 
 									   _("Application Launch Failure!"),
 									   msg,
-									   TRUE);
+									   TRUE,
+									   FALSE);
 					g_free (msg);
 				}
 			}
@@ -393,9 +390,8 @@ gpdf_view_link_action_uri (GPdfView *gpdf_view, LinkURI *link_uri)
 
 		gnome_vfs_mime_application_free (app); 
 	}
-	g_free (uris); 
-	gnome_vfs_uri_unref (vfsuri); 
-	delete uri; 
+	g_free (uri); 
+	g_list_free (uris); 
 }
 
 static void
@@ -929,19 +925,29 @@ gpdf_view_bookmark_selected (GPdfView   *gpdf_view,
 {
 	g_return_if_fail (GPDF_IS_VIEW (gpdf_view));
 
-	switch (link_action->getKind ()) {
-	case actionGoTo:
-		gpdf_view_link_action_goto (
-			gpdf_view, dynamic_cast <LinkGoTo *> (link_action));
-		break;
-	case actionNamed:
-		gpdf_view_link_action_named (
-			gpdf_view, dynamic_cast <LinkNamed *> (link_action));
-		break;
-	default:
-		g_message ("Bookmark selection: unimplemented Action");
-		break;
-	}
+	/*
+	 * Title bookmark (root bookmark on page 1) have null link_action:
+	 * just goto page 1
+	 * These appears on some pdf and no bookmarks at all were displayed
+	 * before because of early recursion failure at build time in
+	 * bookmarks view.
+	 */
+	if (!link_action)
+		gpdf_view_goto_page (gpdf_view, 1);
+	else
+		switch (link_action->getKind ()) {
+		case actionGoTo:
+			gpdf_view_link_action_goto (
+				gpdf_view, dynamic_cast <LinkGoTo *> (link_action));
+			break;
+		case actionNamed:
+			gpdf_view_link_action_named (
+				gpdf_view, dynamic_cast <LinkNamed *> (link_action));
+			break;
+		default:
+			g_message ("Bookmark selection: unimplemented Action");
+			break;
+		}
 }
 
 
